#include <string.h>

#include <stdlib.h>
#include <stdio.h>
#include "tx/request.h"

#define THREADS
#ifdef THREADS
#include <pthread.h>
#endif

#ifdef ORACLE
#include <oci.h>
#endif

#ifdef INGRES
#include <iitxxa.h>
#include <iixa.h>
exec sql include sqlca;
#ifdef USETUXSW
extern struct xa_switch_t iibt_switch;
#endif
#endif

#ifdef DB2
#include <sqlcli.h>
extern struct xa_switch_t db2xa_switch_std;
static SQLRETURN check_error(SQLSMALLINT, SQLHANDLE, SQLRETURN, const char*, SQLCHAR*);
#define CHECK_HANDLE( htype, hndl, rc, msg, sql ) \
	if (rc != SQL_SUCCESS) {return check_error(htype,hndl,rc,msg,sql);}

static SQLCHAR CTSQL[] = "CREATE TABLE XEMP (EMPNO integer NOT NULL PRIMARY KEY, ENAME varchar(32))";
static SQLCHAR ISQL[] = "INSERT INTO XEMP VALUES (?, 'Jim')";
static SQLCHAR USQL[] = "UPDATE XEMP SET ENAME='NEW_NAME' WHERE EMPNO=?";
static SQLCHAR DSQL[] = "DELETE FROM XEMP WHERE EMPNO >= ?";
static SQLCHAR SSQL[] = "select EMPNO, ENAME from XEMP WHERE EMPNO >= ?";
#endif

#define userlogc_warn   printf
#define userlogc_debug  printf
#define userlogc_snprintf   snprintf

#ifdef ORACLE
//extern struct xa_switch_t xaosw;
extern struct xa_switch_t xaoswd;
#endif

typedef struct xabranch {
	int (*dbtest)(struct xabranch *br);
	char op;
	char * arg;
	struct xa_switch_t* xasw;
	int suspended;
	const char* conn_name;
	const char* openstr;
	int rmid;
	XID xid;
} xabranch_t;

static int testDb2(struct xabranch *br);
static int testIng(struct xabranch *br);
static int testOra(struct xabranch *br);

static xabranch_t branches[] = {
#ifdef DB2
	{testDb2, 0, 0, &db2xa_switch_std, 0, "BTDB1", "SREG=T,DB=BTDB1,CHAIN_END=F,TOC=P", 10},
/*	{testDb2, 0, 0, &db2xa_switch_std, 0, "BTDB2", "SREG=T,DB=BTDB2,CHAIN_END=F,TOC=P", 11},*/
#endif
#ifdef INGRES
#ifdef USETUXSW
	{testIng, 0, 0, &iibt_switch, 0, "ING_BTDB1", "INGRES BTDB1 as ING_BTDB1 with tmgroup xyz", 20},
#else
	{testIng, 0, 0, &iixa_switch, 0, "ING_BTDB1", "INGRES BTDB1 as ING_BTDB1", 20},
/*	{testIng, 0, 0, &iixa_switch, 0, "ING_bTDB2", "INGRES BTDB2 as ING_BTDB2", 21},*/
#endif
#endif
#ifdef ORACLE
	{testOra, 0, 0, &xaoswd, 0, "ORCL", "Oracle_XA+Acc=P/scott/tiger+SesTm=60+DbgFl=15+SqlNet=ORCL+DB=ORCL+LogDir=.", 30},
#endif
	{0, 0, 0, NULL}
};

// generate an arbitary but unique XID
static long nextTm = 0L;

static void new_xid(XID *xid)
{
	int i ;

	nextTm = (nextTm == 0L ? (long) time(NULL) : nextTm + 1);

	memset(xid, 0, sizeof(xid));
	xid->formatID = 1L;

	xid->gtrid_length = 4;
	for (i = 0; i < 4; i++)
		xid->data[i] = i;

	xid->bqual_length = 4;
	for (i = 4; i < 8; i++)
		xid->data[i] = (nextTm << ((i - 4) * 8));
}
int fail(const char *reason, int ret)
{
	userlogc_debug("%s: %d\n", reason, ret);
	exit(ret);
}

static int xacall(xabranch_t* br, const char *op, int flags) {
	int ret = XA_OK;
	XID *xid = &(br->xid);
	int rmid = br->rmid;

	userlogc_debug("xaop: %s xa_%s (0x%x)\n", br->conn_name, op, flags);

	if (strcmp(op, "open") == 0)
		ret = (br->xasw->xa_open_entry)((char *) br->openstr, rmid, flags);
	else if (strcmp(op, "close") == 0)
		ret = (br->xasw->xa_close_entry)((char *) "" , rmid, flags);
	else if (strcmp(op, "start") == 0)
		ret = (br->xasw->xa_start_entry)(xid, rmid, flags);
	else if (strcmp(op, "end") == 0)
		ret = (br->xasw->xa_end_entry)(xid, rmid, flags);
	else if (strcmp(op, "prepare") == 0)
		ret = (br->xasw->xa_prepare_entry)(xid, rmid, flags);
	else if (strcmp(op, "commit") == 0)
		ret = (br->xasw->xa_commit_entry)(xid, rmid, flags);
	else if (strcmp(op, "rollback") == 0)
		ret = (br->xasw->xa_rollback_entry)(xid, rmid, flags);

	if (ret != XA_OK) {
		userlogc_debug("%s NOT OK: %d\n", op, ret);
	} else {
//		userlogc_debug("%s OK: %d\n", op, ret);
	}

	return ret;
}

int startTx()
{
#ifdef BLACKTIE
	int ret;
	if ((ret = tx_open()) != TX_OK)
		return fail("tx open err", ret);
	if ((ret = tx_begin()) != TX_OK)
		return fail("tx begin err", ret);
#else
	int i;

	for (i = 0; branches[i].xasw != NULL; i++) {
		xabranch_t *br = &branches[i];
		new_xid(&(br->xid));
	}
#endif
}

static void *finish_tx(void *arg) {
	int i;
	int *commit = (int *) arg;

	for (i = 0; branches[i].xasw != NULL; i++) {
		if (*commit) {
			(void) xacall(&branches[i], "prepare", TMNOFLAGS);
			(void) xacall(&branches[i], "commit", TMNOFLAGS);
		} else {
			(void) xacall(&branches[i], "rollback", TMNOFLAGS);
		}
	}

	return NULL;
}

int terminateTx(int commit, int newThread)
{
	int ret;
//	userlogc_debug("terminating transaction\n");
#ifdef BLACKTIE
	ret = (commit ? tx_commit() : tx_rollback());
	if (ret != TX_OK)
		return fail("tx complete err", ret);
	if ((ret = tx_close()) != TX_OK)
		return fail("tx open err", ret);
#else
	int i;

#ifndef THREADS
	newThread = 0;
#endif

	if (newThread) {
#ifdef THREADS
		userlogc_debug("terminating in a separate thread\n");
		pthread_t thread;
		int rc = pthread_create(&thread, NULL, finish_tx, (void *) &commit);
		pthread_join(thread, NULL);
#endif
	} else {
		userlogc_debug("terminating in the same thread\n");
		finish_tx((void *) &commit);
	}
#endif
	userlogc_debug("transaction terminated\n");
}

static const char *strsub(const char *src, char *dest, size_t sz, const char *match, char *rep)
{
	char *s;
	size_t len;

	if ((s = strstr(src, match)) == NULL)
		return NULL;

	len = s - src;  /* length of the string upto the match */
	strncpy(dest, src, len);
	sprintf(dest+len, "%s%s", rep, s + strlen(match));

	return dest;
}

#ifdef INGRES
static int isInTxn() {
	exec sql begin declare section;
		int trxflag;
	exec sql end declare section;
	exec sql inquire_sql (:trxflag = transaction);
	return trxflag != 0;
}

static int getIngError(char *errbuf, size_t sz) {
	exec sql begin declare section;
		char errmsg[101];
	exec sql end declare section;
	exec sql inquire_sql (:errmsg = ERRORTEXT);

	/*
 	 *- defines for sqlcodes are defined in generr.h
 	 * - the meaning of the codes in documented in SQLRef guide
 	 */
	if (sqlca.sqlcode < 0) {
		userlogc_debug("ing_access: %s\n", errmsg);

		if (errbuf)
			userlogc_snprintf(errbuf, sz, "%s", errmsg);
	} else if (sqlca.sqlcode > 0) {
		userlogc_debug("ing_access: %d %s\n", sqlca.sqlcode, errmsg);
	}

	return sqlca.sqlcode;
}

static void
doIngInsert(int argempnum, char *argename) {
	exec sql begin declare section;
		int empnum = argempnum;
		char *ename = argename;
	exec sql end declare section;

	userlogc_debug("ING: insert into XEMP (empno, ename) values (%d, %s) (inTx=%d) (sqlcode=%d)\n",
		empnum, ename, isInTxn(), sqlca.sqlcode);
	exec sql insert into XEMP (empno, ename) values (:empnum, :ename);
	userlogc_debug("ING: finished insert (inTx=%d) (sqlcode=%d)\n", isInTxn(), sqlca.sqlcode);
}

static void
doIngDelete(int argempnum, char *argename) {
int rcnt;
	exec sql begin declare section;
		int empno = argempnum;
	exec sql end declare section;

	userlogc_debug("ING: delete from XEMP where empno >= %d (inTx=%d)\n", empno, isInTxn());
	exec sql delete from XEMP where empno >= (:empno);
	userlogc_debug("ING: delete finished (inTx=%d)\n", isInTxn());
}


static void
doIngSelect(int argempnum, char *argename, int *rcnt) {
	int validCsr = 0;
	exec sql begin declare section;
		int empno = argempnum;
		char ename[21];
	exec sql end declare section;

	exec sql declare empcsr cursor for
		select empno, ename
		from xemp
		where empno >= :empno;

	exec sql open empcsr;

	if (sqlca.sqlcode != 0) {
		userlogc_debug("ING: doSelect cursor open error %d\n", getIngError(NULL, 0));
		return;
	}

	while (1) {
		exec sql fetch empcsr into :empno, :ename;

		if (sqlca.sqlcode == 0) {
			userlogc_debug("ING: doSelect selected empno %d\n", empno);
			validCsr = 1;
			*rcnt += 1;
			userlogc_debug("%d, %s\n", empno, ename);
		} else {
			userlogc_debug("ING: doSelect selected %d records\n", *rcnt);
			break;
		}
	}
close_cursor:
	userlogc_debug("ING: doSelect close_cursor isValid=%d\n", validCsr);
	if (validCsr)
		exec sql close empcsr;
	userlogc_debug("ING: doSelect cursor closed\n");
}

static int testIng(struct xabranch *br) {
	char op = br->op;
	char *arg = br->arg;
	int cnt = 0;
	exec sql begin declare section;
		char *dbname = (char *) br->conn_name;
	exec sql end declare section;

	userlogc_debug("ING: op=%c data=%s db=%s\n", op, arg, br->conn_name);
	exec sql set connection :dbname;

	if (op == 'i') {	// insert
//		(void) strsub((const char*) ISQL, buf1, sizeof (buf1), "?", empno);
		doIngInsert(atoi(arg), "Jim");
	} else if (op == 's') {	// select
//		(void) strsub((const char*) SSQL, buf1, sizeof (buf1), "?", empno);
		doIngSelect(atoi(arg), "", &cnt);
//		userlogc_snprintf(resp->data, sizeof (resp->data), "%d", rcnt);
	} else if (op == 'u') {	// update
//		(void) strsub((const char*) USQL, buf1, sizeof (buf1), "?", empno);
//		doIngUpdate(atoi(arg), "UpdatedJim");
	} else if (op == 'd') {	// delete
//		(void) strsub((const char*) DSQL, buf1, sizeof (buf1), "?", empno);
		doIngDelete(atoi(arg), "");
	}

	getIngError(NULL,0);

	exec sql set connection NONE;
//exec sql DISCONNECT ALL;
	return 0;
}
#else
static int testIng(struct xabranch *br) {
	return 0;
}
#endif

#ifdef ORACLE
static void get_ora_error(dvoid *errhp, sword status) {
    if  (status == OCI_ERROR) {
        text buf[256];
        sb4 err = 0;
        (void) OCIErrorGet(errhp, (ub4) 1, (text *) NULL, &err, (text *) buf, (ub4) sizeof (buf), OCI_HTYPE_ERROR);

        userlogc_debug("ORA: %s", buf);
    }
}

static int doOraSelect(OCISvcCtx *svcCtx, OCIStmt *stmthp, OCIError *errhp, int empno, int *rcnt) {
    OCIBind *bndp = (OCIBind *) 0;
    text *sql = (text *) "SELECT ENAME,JOB FROM EMP WHERE EMPNO >= :1" ;
    char emp[20];
    char job[20];

    OCIDefine *stmtdef1 = (OCIDefine *) 0;
    OCIDefine *stmtdef2 = (OCIDefine *) 0;

    sword status = OCIStmtPrepare(stmthp, errhp, (text *) sql, (ub4) strlen((char *) sql), (ub4) OCI_NTV_SYNTAX, (ub4) OCI_DEFAULT);

    // THIS IS A C FILE SO DECLARATIONS UP TOP
    userlogc_debug( "ORA: %s :1=%d", sql, empno);

    /* bind empno to the statement */
    if (status == OCI_SUCCESS)
        status = OCIBindByPos(stmthp, &bndp, errhp, 1, (dvoid *) &empno, (sword) sizeof (empno),
                SQLT_INT, (dvoid *) 0, (ub2 *) 0, (ub2 *) 0, (ub4) 0, (ub4 *) 0, OCI_DEFAULT);

    /* define the output variables for the select */
    if (status == OCI_SUCCESS)
        status = OCIDefineByPos(stmthp, &stmtdef1, errhp, 1, (dvoid *)&emp,(sword) sizeof (emp), SQLT_STR,
            (dvoid *) 0, (ub2 *)0, (ub2 *)0, (ub4) OCI_DEFAULT);

    if (status == OCI_SUCCESS)
        status = OCIDefineByPos(stmthp, &stmtdef2, errhp, 2, (dvoid *)&job,(sword) sizeof (job), SQLT_STR,
            (dvoid *) 0, (ub2 *)0, (ub2 *)0, (ub4) OCI_DEFAULT);

    /* exectute the select */
    if (status == OCI_SUCCESS)
        status = OCIStmtExecute(svcCtx, stmthp, errhp, (ub4) 0, (ub4) 0, (CONST OCISnapshot *) NULL, (OCISnapshot *) NULL, OCI_DEFAULT);

    *rcnt = 0;
    if (status != OCI_SUCCESS && status != OCI_NO_DATA) {
        get_ora_error(errhp, status);
        return status;
    } else {
        while (1) {
            status = OCIStmtFetch(stmthp, errhp, (ub4) 1, (ub4) OCI_FETCH_NEXT, (ub4) OCI_DEFAULT);
            if (status != OCI_SUCCESS && status != OCI_SUCCESS_WITH_INFO)
                break;
            userlogc_debug("\tName: %s Job: %s", emp, job);
            (*rcnt) += 1;
        }
        userlogc_debug("\tresult: %d", *rcnt);

        return OCI_SUCCESS;
    }
}

/* execute an SQL statement for the given service context */
static int doOraSql(OCISvcCtx *svcCtx, OCIStmt *stmthp, OCIError *errhp, text *sql, int empno) {
    OCIBind *bndp = (OCIBind *) 0;
    sword status = OCIStmtPrepare(stmthp, errhp, (text *) sql, (ub4) strlen((char *)sql),
        (ub4) OCI_NTV_SYNTAX, (ub4) OCI_DEFAULT);

    userlogc_debug("ORA: %s :1=%d", sql, empno);

    /* bind empno to the statement */
    if (empno > 0 && status == OCI_SUCCESS)
        status = OCIBindByPos(stmthp, &bndp, errhp, 1, (dvoid *) &empno, (sword) sizeof (empno),
                SQLT_INT, (dvoid *) 0, (ub2 *) 0, (ub2 *) 0, (ub4) 0, (ub4 *) 0, OCI_DEFAULT);

    if (status == OCI_SUCCESS)
        status = OCIStmtExecute(svcCtx, stmthp, errhp, (ub4) 1, (ub4) 0,
                                (CONST OCISnapshot *) NULL,
                                (OCISnapshot *) NULL, OCI_DEFAULT);

    if (status != OCI_SUCCESS)
        get_ora_error(errhp, status);

    return status;

}

static int doOraUpdate(OCISvcCtx *svcCtx, OCIStmt *stmthp, OCIError *errhp, int empno) {
    text *sql = (text *) "UPDATE EMP SET JOB='DIRECTOR' WHERE EMPNO=:1" ;
    return doOraSql(svcCtx, stmthp, errhp, sql, empno);
}
static int doOraInsert(OCISvcCtx *svcCtx, OCIStmt *stmthp, OCIError *errhp, int empno) {
    text *sql = (text *) "INSERT INTO EMP (EMPNO,ENAME,JOB,MGR,HIREDATE,SAL,COMM,DEPTNO)"
        " VALUES (:1,'Jim','Janitor','7902','17-DEC-80','900','0','20')";
    return doOraSql(svcCtx, stmthp, errhp, sql, empno);
}
static int doOraDelete(OCISvcCtx *svcCtx, OCIStmt *stmthp, OCIError *errhp, int empno) {
    text *sql = (text *) "DELETE FROM EMP WHERE EMPNO >= :1" ;
    return doOraSql(svcCtx, stmthp, errhp, sql, empno);
}

static int doOraWork(char op, char *arg, OCISvcCtx *svcCtx, OCIStmt *stmthp, OCIError *errhp) {
    sword status = OCI_SUCCESS;
	char buf[512];
    int empno = (*arg ? atoi(arg) : 8000);
#if 1
    if (op == 'i') {
        status = doOraInsert(svcCtx, stmthp, errhp, empno);
    } else if (op == 's') {
        int rcnt = 0;   // no of matching records
        status = doOraSelect(svcCtx, stmthp, errhp, empno, &rcnt);
//        userlogc_snprintf(resp->data, sizeof (resp->data), "%d", rcnt);
    } else if (op == 'u') {
        status = doOraUpdate(svcCtx, stmthp, errhp, empno);
    } else if (op == 'd') {
        status = doOraDelete(svcCtx, stmthp, errhp, empno);
    }

    if (status != OCI_SUCCESS)
        get_ora_error(errhp, status);

    return status;

#else
	text *ISQL = (text *) "INSERT INTO emp "
		" (empno, ename, job, sal, deptno)"
		" VALUES (?,'Jim','Janitor',2400,20)";

	(void) strsub((const char*) ISQL, buf, sizeof (buf), "?", empno);
	text *stmt = (text *) buf;

	status = OCIStmtPrepare(stmthp, errhp, (text *) stmt,
				(ub4) strlen(stmt), (ub4) OCI_NTV_SYNTAX, (ub4) OCI_DEFAULT);

	if (status != OCI_SUCCESS)
		return fail("insert failed", status);

	status = OCIStmtExecute(svcCtx, stmthp, errhp,
				(ub4) 1, (ub4) 0, (CONST OCISnapshot *) NULL, (OCISnapshot *) NULL, OCI_DEFAULT);

	if (status != OCI_SUCCESS) {
		(void) printf("Insert error: %d\n", (int) status);
		return -1;
	}

	return 0;
#endif
}

static int testOra(struct xabranch *br) {
	char op = br->op;
	char *arg = br->arg;
	OCIStmt *stmthp;
	OCIError *errhp;
	OCIEnv *xaEnv;
	OCISvcCtx *svcCtx;

	userlogc_debug("ORA: op=%c data=%s db=%s\n", op, arg, br->conn_name);
	/* opening an XA connection creates an environment and service context */
	xaEnv = (struct OCIEnv *) xaoEnv((text *) conn_name) ;
	svcCtx = (struct OCISvcCtx *) xaoSvcCtx((text *) conn_name);

	if (xaEnv == NULL)
		return fail("ORA:- Unable to obtain env!", -1);

	if (svcCtx == NULL)
		return fail("ORA:- Unable to obtain service context!", -1);

	/* initialise OCI handles */
	if (OCI_SUCCESS != OCIHandleAlloc((dvoid *)xaEnv, (dvoid **)&errhp,
		OCI_HTYPE_ERROR, 0, (dvoid **)0))
		return fail("ORA:- Unable to allocate statement handle", -1);

	if (OCI_SUCCESS != OCIHandleAlloc((dvoid *)xaEnv, (dvoid **)&stmthp,
		OCI_HTYPE_STMT, 0, (dvoid **)0))
		return fail("ORA:- Unable to allocate error handle", -1);

	doOraWork(op, arg, svcCtx, stmthp, errhp);
	return 0;
}
#else
static int testOra(struct xabranch *br) {
	return 0;
}
#endif

#ifdef DB2
/*
 * convert error code from last CLI invokation to a string
 * and use SQLGetDiagRec() to display the SQLSTATE and message
 * for the given handle type
 */
static SQLRETURN check_error( SQLSMALLINT htype, /* A handle type */
					   SQLHANDLE   hndl,  /* A handle */
					   SQLRETURN   rc,   /* Return code */
					   const char* msg,   /* prefix message */
					   SQLCHAR* sql)   /* the sql for statement handles */
{
	SQLCHAR  buffer[SQL_MAX_MESSAGE_LENGTH + 1] ;
	SQLCHAR  sqlstate[SQL_SQLSTATE_SIZE + 1] ;
	SQLINTEGER  sqlcode ;
	SQLSMALLINT length, i = 1;

	switch (rc) {
	case SQL_SUCCESS:   /* 0 */
		return rc;
	case SQL_SUCCESS_WITH_INFO: /* 1 */
		userlogc_warn("DB2: %s error:  SQL_SUCCESS_WITH_INFO: ...", msg);
		break;
	case SQL_NO_DATA_FOUND: /* 100 */
		userlogc_warn("DB2: %s error: SQL_NO_DATA_FOUND: ...", msg);
		break;
	case SQL_ERROR: /* -1 */
		userlogc_warn("DB2: %s error: SQL_ERROR: ...", msg);
		break;
	case SQL_INVALID_HANDLE:	/* -2 */
		userlogc_warn("DB2: %s error: SQL_INVALID HANDLE: ...", msg);
		break;
	default:
		/* SQL_NEED_DATA, SQL_NO_DATA, SQL_STILL_EXECUTING */
		userlogc_warn("DB2: %s error: code=%d: ...", msg, rc);
		break;
	}

	/*
	 * use SQLGetDiagRec() to display SQLSTATE and message for the given handle type
	 */
	while (SQLGetDiagRec(htype, hndl, i++, sqlstate, &sqlcode, buffer,
		SQL_MAX_MESSAGE_LENGTH + 1, &length) == SQL_SUCCESS ) {
	   userlogc_warn( "\t*** SQLSTATE: %s", sqlstate) ;
	   userlogc_warn( "\t*** Native Error Code: %ld", sqlcode) ;
	   userlogc_warn( "\t*** buffer: %s", buffer) ;
	}

	return rc;
}

static SQLRETURN db2init(char *dbalias, SQLHENV *henv, SQLHDBC *hdbc)
{
	SQLHSTMT hstmt; /* statement handle */
	SQLRETURN rc;

	*henv = 0;
	rc = SQLAllocEnv(henv);
	CHECK_HANDLE(SQL_HANDLE_ENV, *henv, rc, "SQLAllocEnv", (SQLCHAR *) "");

	rc = SQLAllocConnect(*henv, hdbc);
	CHECK_HANDLE(SQL_HANDLE_DBC, *hdbc, rc, "SQLAllocConnect", (SQLCHAR *) "");

	rc = SQLConnect(*hdbc, (SQLCHAR *) dbalias, SQL_NTS, NULL, SQL_NTS, NULL, SQL_NTS);
	CHECK_HANDLE(SQL_HANDLE_DBC, *hdbc, rc, "SQLConnect", (SQLCHAR *) "");
	userlogc_debug("DB2: connected to %s\n", dbalias);

	rc = SQLAllocHandle(SQL_HANDLE_STMT, *hdbc, &hstmt);
	CHECK_HANDLE(SQL_HANDLE_STMT, hstmt, rc, "SQLAllocHandle", (SQLCHAR *) "");

	(void) SQLExecDirect(hstmt, CTSQL, SQL_NTS);
	/*CHECK_HANDLE(SQL_HANDLE_STMT, hstmt, rc, "SQLExecDirect", CTSQL);*/

	(void) SQLFreeStmt(hstmt, SQL_DROP);

	return rc;
}

static SQLRETURN db2fini(SQLHENV henv, SQLHDBC hdbc)
{
	/* clean up - free the connection and environment handles */
	SQLDisconnect(hdbc);
	SQLFreeConnect(hdbc);
	SQLFreeEnv(henv);

	return SQL_SUCCESS;
}

static SQLRETURN doDb2Sql(SQLHENV henv, SQLHDBC hdbc, SQLHSTMT shdl, SQLCHAR sql[])
{
	SQLRETURN rc;

	userlogc_debug("DB2: %s\n", sql);
	rc = SQLPrepare(shdl, sql, SQL_NTS);
	CHECK_HANDLE(SQL_HANDLE_STMT, shdl, rc, "SQLPrepare", sql);
	SQLExecDirect(shdl, sql, SQL_NTS);
	CHECK_HANDLE(SQL_HANDLE_STMT, shdl, rc, "SQLExecDirect", sql);

	return SQL_SUCCESS;
}

static SQLRETURN doDb2Select(SQLHENV henv, SQLHDBC hdbc, SQLHSTMT shdl, SQLCHAR sql[], int *rcnt)
{
	SQLCHAR name[33];
	SQLINTEGER empno, col2sz, col1sz;
	SQLRETURN rc = SQLExecDirect(shdl, sql, SQL_NTS);

	userlogc_debug("DB2: %s\n", sql);
	CHECK_HANDLE(SQL_HANDLE_STMT, shdl, rc, "SQLExecDirect", sql);

	/* bind empno and name to columns 1 and 2 of the fetch */
	SQLBindCol(shdl, 1, SQL_C_LONG, (SQLPOINTER) &empno, (SQLINTEGER) sizeof (SQLINTEGER), (SQLINTEGER *) &col1sz);
	SQLBindCol(shdl, 2, SQL_C_CHAR, (SQLPOINTER) name, (SQLINTEGER) sizeof (name), &col2sz);

	*rcnt = 0;
	while (SQLFetch(shdl) == SQL_SUCCESS) {
		*rcnt += 1;
		userlogc_debug("\t(%ld,%s)\n", empno, name);
	}

	userlogc_debug("DB2: selected %d records\n", *rcnt);
	return SQL_SUCCESS;
}

static SQLRETURN doDb2Work(char op, char *empno, SQLHENV henv, SQLHDBC hdbc, SQLHSTMT shdl) {
	SQLRETURN status = SQL_ERROR;
	char buf1[512];
	char buf2[512];
	int rcnt = 0;   // no of matching records

//	(resp->data)[0] = 0;

	if (op == 'i') {
		(void) strsub((const char*) ISQL, buf1, sizeof (buf1), "?", empno);
		status = doDb2Sql(henv, hdbc, shdl, (SQLCHAR *) buf1);
	} else if (op == 's') {
		(void) strsub((const char*) SSQL, buf1, sizeof (buf1), "?", empno);
		status = doDb2Select(henv, hdbc, shdl, (SQLCHAR *) buf1, &rcnt);
//		userlogc_snprintf(resp->data, sizeof (resp->data), "%d", rcnt);
	} else if (op == 'u') {
		(void) strsub((const char*) USQL, buf1, sizeof (buf1), "?", empno);
		status = doDb2Sql(henv, hdbc, shdl, (SQLCHAR *) buf1);
	} else if (op == 'd') {
		(void) strsub((const char*) DSQL, buf1, sizeof (buf1), "?", empno);
		status = doDb2Sql(henv, hdbc, shdl, (SQLCHAR *) buf1);
	}

	return status;
}

static int testDb2(struct xabranch *br) {
	char op = br->op;
	char *arg = br->arg;
	SQLHENV henv;   /* environment handle */
	SQLHDBC hdbc;   /* connection handle */
	SQLRETURN rc = SQL_ERROR;
	SQLHSTMT shdl;

	userlogc_debug("DB2: op=%c data=%s db=%s\n", op, arg, br->conn_name);

	if (db2init((char *) br->conn_name, &henv, &hdbc) != SQL_SUCCESS)
		return (int) rc;

	rc = SQLAllocHandle(SQL_HANDLE_STMT, hdbc, &shdl);

	if ( rc != SQL_SUCCESS ) {
		(void) check_error(SQL_HANDLE_STMT, shdl, rc, "SQLAllocHandle", (SQLCHAR *) "statement handle");
	} else {
		rc = doDb2Work(op, arg, henv, hdbc, shdl);

		(void) SQLFreeStmt(shdl, SQL_DROP);
	}

	db2fini(henv, hdbc);

	return (int) rc;
}
#else
static int testDb2(struct xabranch *br) {
	return 0;
}
#endif

static void *open_rm(void *arg) {
	int ret;
	xabranch_t *br = (xabranch_t *) arg;

	if ((ret = xacall(br, "open", TMNOFLAGS)))
		fail("xa open err", ret);

	return NULL;
}

static void *doIngresWork(void *arg) {
	xabranch_t *br = (xabranch_t *) arg;
	int ret;

	if ((ret = xacall(br, "start", br->suspended ? TMRESUME : TMNOFLAGS)))
		fail("xa start err", ret);

	br->dbtest(br);

	if ((ret = xacall(br, "end", TMSUSPEND)))
		fail("xa end err", ret);

	return NULL;
}

int test1(char op, char * arg, int commit)
{
	int i, ret;

	(void) startTx();

#ifndef INGRES
	for (i = 0; branches[i].xasw != NULL; i++)
		if ((ret = xacall(&branches[i], "start", TMNOFLAGS)))
			return fail("xa start err", ret);
#endif

	for (i = 0; branches[i].xasw != NULL; i++) {
		xabranch_t *br = &branches[i];

		br->op = op;
		br->arg = arg;

#ifdef INGRES
#ifdef THREADS
		pthread_t thread;
		pthread_create(&thread, NULL, doIngresWork, (void *) br);
		pthread_join(thread, NULL);
		if (op == 'i') {
			userlogc_debug("ANOTHER INSERT\n");
			br->suspended = 1;
			br->arg = "8002";
			pthread_create(&thread, NULL, doIngresWork, (void *) br);
			pthread_join(thread, NULL);
			br->suspended = 0;
		}
#else
		doIngresWork(br);
#endif
#else
		br->dbtest(op, arg, br);
#endif
	}

	for (i = 0; branches[i].xasw != NULL; i++)
		(void) xacall(&branches[i], "end", TMSUCCESS);

	(void) terminateTx(commit, 1);  /* commit the tx */
}

static void open_rms() {
	int i, ret;

	for (i = 0; branches[i].xasw != NULL; i++) {
		xabranch_t *br = &branches[i];
		if ((ret = xacall(br, "open", TMNOFLAGS)))
			fail("xa open err", ret);
	}
}

static void close_rms() {
	int i;

	for (i = 0; branches[i].xasw != NULL; i++)
		(void) xacall(&branches[i], "close", TMNOFLAGS);
}

int main(int argc, char *argv[]) {
#ifdef THREADS
	userlogc_debug("PTHREADS: ENABLED\n");
#endif
#ifdef ORACLE
	userlogc_debug("ORA: ENABLED\n");
#endif
#ifdef DB2
	userlogc_debug("DB2: ENABLED\n");
#endif
#ifdef INGRES
	userlogc_debug("INGRES: ENABLED\n");
#endif

	open_rms();

	userlogc_debug("Test delete\n");
	test1('d', "8000", 1);	// delete
	userlogc_debug("Test select\n");
	test1('s', "8000", 1);	// select
	userlogc_debug("Test insert\n");
	test1('i', "8000", 1);	// insert
	userlogc_debug("Test select\n");
	test1('s', "8000", 1);	// select

	close_rms();

	return 0;
}
